#define C86 1
/*#include "stdio.h"*/
#include "chelper.h"

#ifndef C86
#include "ctype.h"
#endif

#define LMAX	512
#define PMAX	256
#define CHAR	1
#define BOL	2
#define EOL	3
#define ANY	4
#define CLASS	5
#define NCLASS	6
#define STAR	7
#define PLUS	8
#define MINUS	9
#define ALPHA	10
#define DIGIT	11
#define NALPHA	12
#define PUNCT	13
#define RANGE	14
#define ENDPAT	15
int	debug = 0;
char	*pp;
char	lbuf[LMAX];
char	pbuf[PMAX];



compile(source)
char *source;		 /* Pattern to compile			 */

/*
 * Compile the pattern into global pbuf[]
 */

{
	register char		*s;	/* Source string pointer	*/
	register char		*lp;	/* Last pattern pointer 	*/
	register int		c;	/* Current character		*/
	int			o;	/* Temp 			*/
	char			*spp;	/* Save beginning of pattern	*/
	char		*cclass();	/* Compile class routine	*/
	s = source;
	if (debug)
		printf("Pattern = \"%s\"\n", s);
	pp = pbuf;
	while (c = *s++)
		{
		/*
						 * STAR, PLUS and MINUS are special.
						 */

		if (c == '*' || c == '+' || c == '-')
			{
			if (pp == pbuf || (o=pp[-1]) == BOL ||
				o == EOL || o == STAR ||
				o == PLUS || o == MINUS)
				badpat("Illegal occurrence op.", source, s);
			store(ENDPAT);
			store(ENDPAT);
			spp = pp;		/* Save pattern end	*/
			while (--pp > lp)	/* Move pattern down	*/
				*pp = pp[-1];	/* one byte		*/
			*pp =	((c == '*') ? STAR : (c == '-' ? MINUS : PLUS));
			pp = spp;		/* Restore pattern end	*/
			continue;
			}

		/*
				 *	 All the rest.
				 */

		lp = pp;			/*	Remember start	*/
		switch(c)
			{
		case '^':
			store(BOL);
			break;
		case '$':
			store(EOL);
			break;
		case '.':
			store(ANY);
			break;
		case '[':
			s = cclass(source, s);
			break;
		case ':':
			if (*s)
				{
				c = *s++;
				switch(tolower(c))
					{
				case 'a':
					store(ALPHA);
					break;
				case 'd':
					store(DIGIT);
					break;
				case 'n':
					store(NALPHA);
					break;
				case '_':
					store(PUNCT);
					break;
				default:
					badpat("Unknown : type", source, s);
					}
				break;
				}
			else	badpat("No : type", source, s);
		case '\\':
			if (*s)
				c = *s++;
		default:
			store(CHAR);
			store(tolower(c));
			}
		}
	store(ENDPAT);
	store(0);				/*   Terminate string	  */
	if (debug)
		{
		for (lp = pbuf; lp < pp;)
			{
			if ((c = (*lp++ & 0377)) < ' ')
				printf("\\%o ", c);
			else	printf("%c ", c);
			}
		printf("\n");
		}
}

char *cclass(source, src)
char		*source;	/* Pattern start -- for error msg.	*/
char		*src;		/* Class start				*/

/*
 * Compile a class (within [])
 */

{
	register char	*s;		/* Source pointer		*/
	register char	*cp;		/* Pattern start		*/
	register int	c;		/* Current character		*/
	int		o;		/* Temp 			*/
	s = src;
	o = CLASS;
	if (*s == '^')
		{
		++s;
		o = NCLASS;
		}
	store(o);
	cp = pp;
	store(0);				/* Byte count		*/
	while ((c = *s++) && c!=']')
		{
		if (c == '\\')
			{		/* Store quoted char	*/
			if ((c = *s++) == '\0') /* Gotta get something	*/
				badpat("Class terminates badly", source, s);
			else	store(tolower(c));
			}
		else if (c == '-' &&
			(pp - cp) > 1 && *s != ']' && *s != '\0')
			{
			c = pp[-1];		/* Range start		*/
			pp[-1] = RANGE; 	/* Range signal 	*/
			store(c);		/* Re-store start	*/
			c = *s++;		/* Get end char and	*/
			store(tolower(c));	/* Store it		*/
			}
		else
			{
			store(tolower(c));	/* Store normal char	*/
			}
		}
	if (c != ']')
		badpat("Unterminated class", source, s);
	if ((c = (pp - cp)) >= 256)
		badpat("Class too large", source, s);
	if (c == 0)
		badpat("Empty class", source, s);
	*cp = c;
	return(s);
}

store(op)
{
	if (pp >= &pbuf[PMAX])
		abort("Pattern too complex\n");
	*pp++ = op;
}


badpat(message, source, stop)
char		*message;	/* Error message			*/
char		*source;	/* Pattern start			*/
char		*stop;		/* Pattern end				*/
{
	fprintf(stderr, "Bad regular exression\n");
	exit(-1);
}

match()

/*
 * Match the current line (in lbuf[]), return 1 if it does.
 */

{
	register char	*l;		/* Line pointer 		*/
	char *pmatch();
	for (l = lbuf; *l; l++)
		{
		if (pmatch(l, pbuf))
			return(1);
		}
	return(0);
}

char *pmatch(line, pattern)
char		*line;		/* (partial) line to match		*/
char		*pattern;	/* (partial) pattern to match		*/
{
	register char	*l;		/* Current line pointer 	*/
	register char	*p;		/* Current pattern pointer	*/
	register char	c;		/* Current character		*/
	char		*e;		/* End for STAR and PLUS match	*/
	int		op;		/* Pattern operation		*/
	int		n;		/* Class counter		*/
	char		*are;		/* Start of STAR match		*/
	l = line;
	if (debug > 1)
		printf("pmatch(\"%s\")\n", line);
	p = pattern;
	while ((op = *p++) != ENDPAT)
		{
		if (debug > 1)
			printf("byte[%d] = 0%o, '%c', op = 0%o\n",
			l-line, *l, *l, op);
		switch(op)
			{
		case CHAR:
			if (tolower(*l++) != *p++)
				return(0);
			break;
		case BOL:
			if (l != lbuf)
				return(0);
			break;
		case EOL:
			if (*l != '\0')
				return(0);
			break;
		case ANY:
			if (*l++ == '\0')
				return(0);
			break;
		case DIGIT:
			if ((c = *l++) < '0' || (c > '9'))
				return(0);
			break;
		case ALPHA:
			c = tolower(*l++);
			if (c < 'a' || c > 'z')
				return(0);
			break;
		case NALPHA:
			c = tolower(*l++);
			if (c >= 'a' && c <= 'z')
				break;
			else if (c < '0' || c > '9')
				return(0);
			break;
		case PUNCT:
			c = *l++;
			if (c == 0 || c > ' ')
				return(0);
			break;
		case CLASS:
		case NCLASS:
			c = tolower(*l++);
			n = *p++ & 0377;
			do
				{
				if (*p == RANGE)
					{
					p += 3;
					n -= 2;
					if (c >= p[-2] && c <= p[-1])
						break;
					}
				else if (c == *p++)
					break;
				}
			while (--n > 1);
			if ((op == CLASS) == (n <= 1))
				return(0);
			if (op == CLASS)
				p += n - 2;
			break;
		case MINUS:
			e = pmatch(l, p);	/* Look for a match	*/
			while (*p++ != ENDPAT); /* Skip over pattern	*/
			if (e)			/* Got a match? 	*/
				l = e;		/* Yes, update string	*/
			break;			/* Always succeeds	*/
		case PLUS:			/* One or more ...	*/
			if ((l = pmatch(l, p)) == 0)
				return(0);	/* Gotta have a match	*/
		case STAR:			/* Zero or more ...	*/
			are = l;		/* Remember line start	*/
			while (*l && (e = pmatch(l, p)))
				l = e;		/* Get longest match	*/
			while (*p++ != ENDPAT); /* Skip over pattern	*/
			while (l >= are)
				{	/* Try to match rest	*/
				if (e = pmatch(l, p))
					return(e);
				--l;		/* Nope, try earlier	*/
				}
			return(0);		/* Nothing else worked	*/
		default:
			printf("Bad op code %d\n", op);
			abort("Cannot happen -- match\n");
			}
		}
	return(l);
}

